use nom::{
    bytes::complete::tag_no_case,
    character::complete::{hex_digit1, line_ending, not_line_ending, space1},
    combinator::{map_parser, map_res, opt},
    error::{Error, ErrorKind},
    multi::fold_many0,
    sequence::{preceded, separated_pair, terminated},
    IResult,
};

#[derive(Default)]
pub struct InstrumentSet([u8; 32]);

impl InstrumentSet {
    pub fn generate(input: &str) -> Self {
        match Self::parse(input) {
            Ok((_, set)) => set,
            Err(_) => panic!(),
        }
    }

    fn parse<'i>(input: &'i str) -> IResult<&'i str, Self> {
        fold_many0(
            terminated(
                opt(preceded(
                    tag_no_case("#WAVE"),
                    map_parser(not_line_ending, wave_numbers),
                )),
                line_ending,
            ),
            Self::default,
            |mut set, bytes: Option<(usize, u8)>| match bytes {
                Some((slot, sample)) => {
                    let pos = slot - 0x20;
                    set.0[pos] = sample;
                    set
                }
                None => set,
            },
        )(input)
    }
}

fn wave_numbers(input: &str) -> IResult<&str, (usize, u8)> {
    separated_pair(
        map_res(hex_digit1, |s: &str| {
            s.parse::<usize>().map_err(|_| nom::error::make_error::<&str, Error<&str>>(s, ErrorKind::HexDigit)).and_then(|n| {
                if (0x20..0x30).contains(&n) {
                    Ok(n)
                } else {
                    Err(nom::error::make_error(s, ErrorKind::MapRes))
                }
            })
        }),
        space1,
        map_res(hex_digit1, |s: &str| s.parse::<u8>()),
    )(input)
}
